﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Drawing;
using System.Windows.Forms;

namespace GameOfLife
{
    public partial class frmMain : Form
    {
        enum CellState
        {
            Death, Living
        }
        private class Canvas : PictureBox
        {
            public Canvas()
            {
                // Einstellungen nach dem Artikel
                // https://www.mycsharp.de/wbb2/thread.php?threadid=59978
                // Sind aber nicht notwendig! Siehe Paint-Event von Canvas
                // Daher wäre die Ableitung nicht erforderlich!
                // Diese wurde nur gemacht, um DoubleBuffered setzen zu können

                //this.DoubleBuffered = true;
                //this.SetStyle(ControlStyles.DoubleBuffer, true);
                //this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
                //this.SetStyle(ControlStyles.CacheText, true);
                //this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
                //this.SetStyle(ControlStyles.UserPaint, true);
            }
        }

        private class RuleSet
        {
            public int[] ruleKeepAlive { get; private set; }
            public int[] ruleBecomeBorn { get; private set; }

            public RuleSet(int[] keepAlive, int[] becomeBorn)
            {
                ruleKeepAlive = keepAlive;
                ruleBecomeBorn = becomeBorn;
            }
        }

        Canvas canvas;

        // Size gridSize;
        Size rectangleSize;
        Size canvasGridSize;

        Color colorLivingCell = Color.MediumBlue;
        Color colorDeathCell = Color.Black;

        int CellsPerRow = 50;
        int CellsPerColumn = 40;

        CellState[,] cells;
        StringCollection scells;

        Timer timer;
        int numberOfTimeSlice = 0;

        List<RuleSet> rules;
        RuleSet actualRuleSet;
        StringCollection rulesettings;
        string actualRule = "";

        public frmMain()
        {
            InitializeComponent();

            rules = new List<RuleSet>();

            // Einstellungen nach dem Artikel
            // https://www.mycsharp.de/wbb2/thread.php?threadid=59978
            // Sind aber nicht notwendig! Siehe Paint-Event von Canvas

            //this.DoubleBuffered = true;
            //this.SetStyle(ControlStyles.DoubleBuffer, true);
            //this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);
            //this.SetStyle(ControlStyles.CacheText, true);
            //this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            //this.SetStyle(ControlStyles.UserPaint, true);

            canvas = new Canvas();
            // canvas.Size wird in initializeCanvas() berechnet
            canvas.Location = new Point(10, 10);
            canvas.BorderStyle = BorderStyle.FixedSingle;
            canvas.BackColor = Color.LightGray;
            canvas.Paint += canvas_Paint;
            canvas.MouseClick += canvas_MouseClick;

            Controls.Add(canvas);

            timer = new Timer();
            timer.Enabled = false;
            timer.Tick += nextTimeSlice;

            readSettings();
            insertRuleSetInCombobox();
            initializeArrays();
            initializeCanvas();

            // Steuerelementwerte aktualisieren
            numericUpDownZeitintervall.Value = timer.Interval;
            numericUpDownCellsPerRow.Value = CellsPerRow;
            numericUpDownCellsPerColumn.Value = CellsPerColumn;
            labelColorDeathCell.BackColor = colorDeathCell;
            labelColorLivingCell.BackColor = colorLivingCell;

            if (comboBoxRuleSet.Items.Count > 0)
            {
                int index = 0;
                foreach (string rule in comboBoxRuleSet.Items)
                {
                    if (rule == actualRule)
                    {
                        comboBoxRuleSet.SelectedIndex = index;
                        actualRuleSet = rules[index];
                        break;
                    }
                    index++;
                }
            }
            // MessageBox.Show(this.ClientSize.Width.ToString() + "\n" + this.ClientSize.Height.ToString());
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (checkBoxWriteAtProgrammEnd.Checked)
            {
                saveCellState();
                writeSettings();
            }
            else
            {
                DialogResult dr = MessageBox.Show("Einstellungen sichern?", "Programmende", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (dr == DialogResult.Yes)
                {
                    saveCellState();
                    writeSettings();
                }
            }

            frmAboutBox a = new frmAboutBox();
            a.ShowDialog();
        }

        private void readSettings()
        {
            colorDeathCell = Properties.Settings.Default.ColorDeathCell;
            colorLivingCell = Properties.Settings.Default.ColorLivingCell;
            timer.Interval = Properties.Settings.Default.TimeSlice;

            CellsPerRow = Properties.Settings.Default.CellsPerRow;
            CellsPerColumn = Properties.Settings.Default.CellsPerColumn;

            scells = Properties.Settings.Default.LastImage;

            numberOfTimeSlice = Properties.Settings.Default.Generation;
            checkBoxWriteAtProgrammEnd.Checked = Properties.Settings.Default.SaveAtEnd;
            actualRule = Properties.Settings.Default.ActualRule;

            this.Size = Properties.Settings.Default.FormSize;
            this.WindowState = (FormWindowState)Properties.Settings.Default.WindowState;
            this.Location = Properties.Settings.Default.WindowPosition;

            rulesettings = Properties.Settings.Default.RuleSettings;

            // Gelesene Regeln in RuleSet-Objekte "gießen"
            try
            {
                foreach (string rule in rulesettings)
                {
                    string[] readRuleSetting = rule.Split('/');
                    int[] rule1 = new int[readRuleSetting[0].Length];
                    for (int i = 0; i < readRuleSetting[0].Length; i++)
                    {
                        int createNumber = Int32.Parse(readRuleSetting[0][i].ToString());
                        rule1[i] = createNumber;
                    }

                    int[] rule2 = new int[readRuleSetting[1].Length];
                    for (int i = 0; i < readRuleSetting[1].Length; i++)
                    {
                        int createNumber = Int32.Parse(readRuleSetting[1][i].ToString());
                        rule2[i] = createNumber;
                    }

                    RuleSet rs = new RuleSet(rule1, rule2);
                    rules.Add(rs);
                }
            }
            catch (ArgumentOutOfRangeException ex)
            {
                MessageBox.Show(ex.Message,
                                "Fehler bei Lesen der Regelsets",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
            }
            catch (FormatException ex)
            {
                MessageBox.Show(ex.Message,
                                "Fehler bei Format eines Regelsets",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
            }
        }

        private void insertRuleSetInCombobox()
        {
            comboBoxRuleSet.Items.Clear();
            foreach (RuleSet rs in rules)
            {
                string rule = "";
                // Bleibt am Leben-Regel
                for (int i = 0; i < rs.ruleKeepAlive.Length; i++)
                {
                    rule += rs.ruleKeepAlive[i].ToString();
                }
                rule += "/";
                // Wird neu geboren-Rregel
                for (int i = 0; i < rs.ruleBecomeBorn.Length; i++)
                {
                    rule += rs.ruleBecomeBorn[i].ToString();
                }

                comboBoxRuleSet.Items.Add(rule);
            }
        }
        private void writeSettings()
        {
            Properties.Settings.Default.ColorDeathCell = colorDeathCell;
            Properties.Settings.Default.ColorLivingCell = colorLivingCell;
            Properties.Settings.Default.TimeSlice = timer.Interval;
            Properties.Settings.Default.CellsPerRow = CellsPerRow;
            Properties.Settings.Default.CellsPerColumn = CellsPerColumn;
            Properties.Settings.Default.LastImage = scells;
            Properties.Settings.Default.Generation = numberOfTimeSlice;
            Properties.Settings.Default.SaveAtEnd = checkBoxWriteAtProgrammEnd.Checked;
            Properties.Settings.Default.ActualRule = comboBoxRuleSet.Items[comboBoxRuleSet.SelectedIndex].ToString();
            Properties.Settings.Default.WindowState = (int)this.WindowState;

            if (this.WindowState==FormWindowState.Maximized)
            {
                
                this.WindowState = FormWindowState.Normal;
                this.Hide();
            }
            Properties.Settings.Default.FormSize = this.Size;
            Properties.Settings.Default.WindowPosition = this.Location;

            rulesettings.Clear();
            foreach (string rule in comboBoxRuleSet.Items)
            {
                rulesettings.Add(rule);
            }
            Properties.Settings.Default.RuleSettings = rulesettings;

            Properties.Settings.Default.Save();
        }

        private void saveCellState()
        {
            // Aktuellen Stand in scells speichern
            if (scells == null)
                scells = new StringCollection();
            else
                scells.Clear();

            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    if (cells[col, row] == CellState.Living)
                    {
                        scells.Add("1");
                    }
                    else
                    {
                        scells.Add("0");
                    }
                }
            }
        }

        private void initializeArrays()
        {
            // Nach Programmstart oder Ändern der Zellenzahl pro Zeile oder Spalte
            // müssen neue Arrays erzeugt und initialisiert werden
            cells = new CellState[CellsPerRow, CellsPerColumn];

            if ((scells != null) && (scells.Count == cells.Length))
            {
                for (int row = 0; row < CellsPerColumn; row++)
                {
                    for (int col = 0; col < CellsPerRow; col++)
                    {
                        if (scells[col + row * CellsPerRow] == "1")
                            cells[col, row] = CellState.Living;
                        else
                            cells[col, row] = CellState.Death;
                    }
                }
            }
        }
        private void initializeCanvas()
        {
            //canvasGridSize = new Size(canvas.Size.Width / CellsPerRow,
            //                          canvas.Size.Height / CellsPerColumn
            //                         );

            //rectangleSize = new Size(canvas.Size.Width / CellsPerRow - 1,
            //                         canvas.Size.Height / CellsPerColumn - 1
            //                        );

            canvasGridSize = new Size((this.ClientSize.Width - 18) / CellsPerRow,
                                      (this.ClientSize.Height - 170) / CellsPerColumn
                                     );

            rectangleSize = new Size(canvasGridSize.Width - 1,
                                     canvasGridSize.Height - 1
                                    );
            // Da die Canvas etwas größer sein kann, als die
            // Anzahl Rechtecke * canvasGridSize,
            // wird die Canvasgröße angepasst
            canvas.Size = new Size(canvasGridSize.Width * CellsPerRow,
                                   canvasGridSize.Height * CellsPerColumn
                                  );
        }

        private void createRandomValues()
        {
            Random rnd = new Random();
            // scells.Clear();

            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    if (Convert.ToBoolean(rnd.Next(0, 2)))
                    {
                        cells[col, row] = CellState.Living;
                    }
                    else
                    {
                        cells[col, row] = CellState.Death;
                    }
                }
            }
        }
        private void nextTimeSlice(object sender, EventArgs e)
        {
            // Regeln implementieren
            // siehe de.wikipedia.org/wiki/Conways_Spiel_des_Lebens#Die_Spielregeln
            // 1. Eine tote Zelle mit genau drei lebenden Nachbarn wird in der Folgegeneration neu geboren.
            // 2. Lebende Zellen mit weniger als zwei lebenden Nachbarn sterben in der Folgegeneration an Einsamkeit.
            // 3. Lebende Zellen mit mehr als drei lebenden Nachbarn sterben in der Folgegeneration an Überbevölkerung.
            // 4. Eine lebende Zelle mit zwei oder drei lebenden Nachbarn bleibt in der Folgegeneration am Leben.
            // Die Regeln werden für alle gleichzeitig angewendet
            // Zweidimensionales Array für alle Zellen mit der Anzahl der Nachbarn als Wert
            int[,] neighboursum = new int[CellsPerRow, CellsPerColumn];

            // Zellen sind von oben nach unten und umgekehrt sowie von links nach rechts
            // und umgekehrt benachbart!
            // Jede Zelle hat 8 Nachbarn in den Himmelsrichtungen N, NO, O, SO, S, SW, W, NW

            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    int neighbours = 0;
                    // Nachbar-Zellen identifizieren
                    int above = row - 1;
                    if (above < 0)
                        above = CellsPerColumn - 1;
                    int below = row + 1;
                    if (below >= CellsPerColumn)
                        below = 0;
                    int left = col - 1;
                    if (left < 0)
                        left = CellsPerRow - 1;
                    int right = col + 1;
                    if (right >= CellsPerRow)
                        right = 0;
                    //Anzahl lebende Nachbarn berechnen für jede benachbarte Zelle
                    if (cells[col, above] == CellState.Living)      //N
                        neighbours++;
                    if (cells[right, above] == CellState.Living)    //NO
                        neighbours++;
                    if (cells[right, row] == CellState.Living)      //O
                        neighbours++;
                    if (cells[right, below] == CellState.Living)    //SO
                        neighbours++;
                    if (cells[col, below] == CellState.Living)      //S
                        neighbours++;
                    if (cells[left, below] == CellState.Living)     //SW
                        neighbours++;
                    if (cells[left, row] == CellState.Living)       //W
                        neighbours++;
                    if (cells[left, above] == CellState.Living)      //NW
                        neighbours++;

                    neighboursum[col, row] = neighbours;
                }
            }

            // Neuen Zellzustand gemäß Regeln festlegen
            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    // Regel 1 (Wird geboren)
                    if (cells[col, row] == CellState.Death)
                    {
                        for (int i = 0; i < actualRuleSet.ruleBecomeBorn.Length; i++)
                        {
                            if (neighboursum[col, row] == actualRuleSet.ruleBecomeBorn[i])
                            {
                                cells[col, row] = CellState.Living;
                                break;
                            }
                        }
                    }
                    else if (cells[col, row] == CellState.Living)
                    {
                        // Erstmal sterben lassen
                        cells[col, row] = CellState.Death;
                        // Regel 4 (Bleibt am Leben)
                        for (int i = 0; i < actualRuleSet.ruleKeepAlive.Length; i++)
                        {
                            if (neighboursum[col, row] == actualRuleSet.ruleKeepAlive[i])
                            {
                                cells[col, row] = CellState.Living;
                                break;
                            }
                        }
                    }
                }
            }

            numberOfTimeSlice++;
            canvas.Invalidate();

            if ((int)numericUpDownCounterGenerations.Value == numberOfTimeSlice)
                buttonStartStopp_Click(sender, e);
        }
        private void buttonStartStopp_Click(object sender, EventArgs e)
        {
            if (timer.Enabled)
            {
                timer.Enabled = false;
                buttonStartStopp.Text = "Starte Game of Life";
                buttonOneTimeSlice.Enabled = true;
                buttonSetAllCellsToDeath.Enabled = true;
                buttonSetAllCellsToLiving.Enabled = true;
                buttonSetAllCellsRandom.Enabled = true;
                buttonNewRule.Enabled = true;
                numericUpDownCellsPerRow.Enabled = true;
                numericUpDownCellsPerColumn.Enabled = true;
                numericUpDownCounterGenerations.Enabled = true;
                comboBoxRuleSet.Enabled = true;
            }
            else
            {
                timer.Enabled = true;
                buttonStartStopp.Text = "Stoppe Game of Life";
                buttonOneTimeSlice.Enabled = false;
                buttonSetAllCellsToDeath.Enabled = false;
                buttonSetAllCellsToLiving.Enabled = false;
                buttonSetAllCellsRandom.Enabled = false;
                buttonNewRule.Enabled = false;
                numericUpDownCellsPerRow.Enabled = false;
                numericUpDownCellsPerColumn.Enabled = false;
                numericUpDownCounterGenerations.Enabled = false;
                comboBoxRuleSet.Enabled = false;
            }
        }

        private void canvas_Paint(object sender, PaintEventArgs e)
        {
            this.Text = "Game of Life (Generation " +
                        numberOfTimeSlice.ToString() +
                        ")";

            // Übersicht zu Artikeln rund ums Zeichnen:
            // https://www.mycsharp.de/wbb2/thread.php?threadid=59980

            // Entscheidender Hinweis, um Flackern zu vermeiden:
            // Man sollte im Paint/OnPaint immer mit e.Graphics zeichnen
            // und nie mit Control.CreateGraphics versuchen selbst
            // ein Graphics-Objekt zu erzeugen.
            // https://www.mycsharp.de/wbb2/thread.php?threadid=21782

            // Zeichen der Rechtecke
            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    Brush brush;

                    if (cells[col, row] == CellState.Death)
                        brush = new SolidBrush(colorDeathCell);
                    else
                        brush = new SolidBrush(colorLivingCell);

                    Point position = new Point(col * canvasGridSize.Width,
                                               row * canvasGridSize.Height
                                              );
                    e.Graphics.FillRectangle(brush, new Rectangle(position, rectangleSize));
                }
            }
        }

        private void canvas_MouseClick(object sender, MouseEventArgs e)
        {
            int col = e.Location.X / (rectangleSize.Width + 1);
            int row = e.Location.Y / (rectangleSize.Height + 1);

            labelX.Text = col.ToString();
            labelY.Text = row.ToString();

            // MessageBox.Show("Col=" + col.ToString() + "\r\n" + "Row=" + row.ToString());

            if (col >= 0 & row >= 0)
            {
                if (cells[col, row] == CellState.Death)
                    cells[col, row] = CellState.Living;
                else
                    cells[col, row] = CellState.Death;

                canvas.Invalidate();

            }
        }
        private void buttonOneTimeSlice_Click(object sender, EventArgs e)
        {
            nextTimeSlice(sender, e);
        }

        private void numericUpDownCellsPerRow_ValueChanged(object sender, EventArgs e)
        {
            CellsPerRow = (int)numericUpDownCellsPerRow.Value;
            initializeArrays();
            initializeCanvas();
            canvas.Invalidate();
        }

        private void numericUpDownCellsPerColumn_ValueChanged(object sender, EventArgs e)
        {
            CellsPerColumn = (int)numericUpDownCellsPerColumn.Value;
            initializeArrays();
            initializeCanvas();
            canvas.Invalidate();
        }

        private void numericUpDownZeitintervall_ValueChanged(object sender, EventArgs e)
        {
            timer.Interval = (int)numericUpDownZeitintervall.Value;
        }

        private void labelColorDeathCell_Click(object sender, EventArgs e)
        {
            ColorDialog cd = new ColorDialog();
            DialogResult dr;
            if ((dr = cd.ShowDialog()) == DialogResult.OK)
            {
                colorDeathCell = cd.Color;
                labelColorDeathCell.BackColor = colorDeathCell;
                canvas.Invalidate();
            }
        }

        private void labelColorLivingCell_Click(object sender, EventArgs e)
        {
            ColorDialog cd = new ColorDialog();
            DialogResult dr;
            if ((dr = cd.ShowDialog()) == DialogResult.OK)
            {
                colorLivingCell = cd.Color;
                labelColorLivingCell.BackColor = colorLivingCell;
                canvas.Invalidate();
            }
        }

        private void buttonSetAllCellsToLiving_Click(object sender, EventArgs e)
        {
            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    cells[col, row] = CellState.Living;
                }
            }
            canvas.Invalidate();
        }

        private void buttonSetAllCellsToDeath_Click(object sender, EventArgs e)
        {
            for (int row = 0; row < CellsPerColumn; row++)
            {
                for (int col = 0; col < CellsPerRow; col++)
                {
                    cells[col, row] = CellState.Death;
                }
            }
            numberOfTimeSlice = 0;
            canvas.Invalidate();
        }

        private void buttonSetAllCellsRandom_Click(object sender, EventArgs e)
        {
            createRandomValues();
            numberOfTimeSlice = 0;
            canvas.Invalidate();
        }

        private void buttonSaveCellState_Click(object sender, EventArgs e)
        {
            saveCellState();
            writeSettings();
        }

        private void comboBoxRuleSet_SelectedIndexChanged(object sender, EventArgs e)
        {
            actualRuleSet = rules[comboBoxRuleSet.SelectedIndex];
        }

        private void buttonNewRule_Click(object sender, EventArgs e)
        {
            // Testen, ob TextBox-Inhalte gültig sind
            int keepAlive;
            int becomeLiving;

            if (!Int32.TryParse(textBoxKeepAlive.Text, out keepAlive))
            {
                MessageBox.Show("Ungültiger Wert bei 'Bleibt lebendig'",
                                "Fehler",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
                textBoxKeepAlive.Text = "";
                return;
            }

            if (textBoxKeepAlive.Text.Contains("9"))
            {
                MessageBox.Show("Ungültiger Wert bei 'Bleibt lebendig'",
                                "Fehler",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
                textBoxKeepAlive.Text = "";
                return;
            }

            if (!Int32.TryParse(textBoxBecomeLiving.Text, out becomeLiving))
            {
                MessageBox.Show("Ungültiger Wert bei 'Wird lebendig'",
                                "Fehler",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
                textBoxBecomeLiving.Text = "";
                return;
            }

            if (textBoxBecomeLiving.Text.Contains("9"))
            {
                MessageBox.Show("Ungültiger Wert bei 'Wird lebendig'",
                                "Fehler",
                                MessageBoxButtons.OK,
                                MessageBoxIcon.Error
                               );
                textBoxBecomeLiving.Text = "";
                return;
            }

            // Okay, Regel hat gültiges Format
            // Nun ein RuleSet-Objekt erzeugen
            string rule = keepAlive.ToString() + "/" + becomeLiving.ToString();

            string[] readRuleSetting = rule.Split('/');
            int[] rule1 = new int[readRuleSetting[0].Length];
            for (int i = 0; i < readRuleSetting[0].Length; i++)
            {
                int createNumber = Int32.Parse(readRuleSetting[0][i].ToString());
                rule1[i] = createNumber;
            }

            int[] rule2 = new int[readRuleSetting[1].Length];
            for (int i = 0; i < readRuleSetting[1].Length; i++)
            {
                int createNumber = Int32.Parse(readRuleSetting[1][i].ToString());
                rule2[i] = createNumber;
            }

            // Gibt es diese Regel schon?
            foreach (string ruletest in comboBoxRuleSet.Items)
            {
                if (ruletest == rule)
                {
                    MessageBox.Show("Diese Regel existiert bereits!",
                                    "Fehler",
                                    MessageBoxButtons.OK,
                                    MessageBoxIcon.Stop
                                   );
                    return;
                }
            }

            // Nein --> Regel aufnehmen in Liste
            RuleSet rs = new RuleSet(rule1, rule2);
            rules.Add(rs);

            comboBoxRuleSet.Items.Add(rule);

            textBoxKeepAlive.Text = "";
            textBoxBecomeLiving.Text = "";

            MessageBox.Show("Regel gespeichert",
                            "Neue Regel",
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Information
                           );

            // Neue Regel aktivieren
            comboBoxRuleSet.SelectedIndex = comboBoxRuleSet.Items.Count - 1;
        }

        private void frmMain_Resize(object sender, EventArgs e)
        {
            Form f = (Form)sender;
            int width = f.ClientSize.Width;
            int height = f.ClientSize.Height;

            groupBoxSettings.Location = new Point(10, height - 160);
            canvas.Size = new Size(width - 23, height - 190);
            initializeCanvas();
            canvas.Invalidate();
        }
    }
}
